#include <math.h>
#include <iostream>
#include <stddef.h>
#include "realtime.h"
#include "jacobi.h"


#define U(j,i) data.afU[((j) - data.iRowFirst) * data.iCols + (i)]
#define F(j,i) data.afF[((j) - data.iRowFirst) * data.iCols + (i)]

using namespace std;

// setting values, init mpi, omp etc
void Init(JacobiData &data, int &argc, char **argv)
{

// default medium
        data.iCols      = 2000;
        data.iRows      = 2000;
        data.fAlpha     = 0.8;
        data.fRelax     = 1.0;
        data.fTolerance = 1e-10;
        data.iIterMax   = 50;
#ifdef READ_INPUT
        cout << "Input n - matrix size in x direction: ";
        cin >> data.iCols;
        cout << "\nInput m - matrix size in y direction: ";
        cin >> data.iRows;
        cout << "\nInput alpha - Helmholtz constant: ";
        cin >> data.fAlpha;
        cout << "\nInput relax - Successive over-relaxation parameter: ";
        cin >> data.fRelax;
        cout << "\nInput tol - error tolerance for iterrative solver: ";
        cin >> data.fTolerance;
        cout << "\nInput mits - Maximum iterations for solver: ";
        cin >> data.iIterMax;
#elif defined DATA_LARGE
        data.iCols      = 7000;
        data.iRows      = 7000;
        data.fAlpha     = 0.8;
        data.fRelax     = 1.0;
        data.fTolerance = 1e-12;
        data.iIterMax   = 2;

#elif defined DATA_SMALL
        data.iCols      = 200;
        data.iRows      = 200;
        data.fAlpha     = 0.8;
        data.fRelax     = 1.0;
        data.fTolerance = 1e-7;
        data.iIterMax   = 1000;
#endif
        cout << "\n-> matrix size: " << data.iCols << "x" << data.iRows
             << "\n-> alpha: " << data.fAlpha
             << "\n-> relax: " << data.fRelax
             << "\n-> tolerance: " << data.fTolerance
             << "\n-> #of iterations: " << data.iIterMax << endl << endl;


    // MPI values, set to defaults to avoid data inconsistency
    data.iMyRank   = 0;
    data.iNumProcs = 1;
    data.iRowFirst = 0;
    data.iRowLast  = data.iRows - 1;

    // memory allocation for serial & omp
    data.afU = new double[data.iRows * data.iCols];
    data.afF = new double[data.iRows * data.iCols];

    // calculate dx and dy
    data.fDx = 2.0 / (data.iCols - 1);
    data.fDy = 2.0 / (data.iRows - 1);

    data.iIterCount = 0;   

    return;
}

// final cleanup routines
void Finish(JacobiData &data)
{
    delete [] data.afU;
    delete [] data.afF;

    return;
}

// print result summary
void PrintResults(const JacobiData &data)
{
    if (data.iMyRank == 0)
    {
        cout << " Number of iterations : " << data.iIterCount << endl;
        cout << " Residual             : " << data.fResidual << endl;
        cout << " Solution Error       : " << data.fError << endl;
        cout << " Elapsed Time         : " << data.fTimeStop - data.fTimeStart << endl;
        cout << " MFlops               : " << 0.000013 * data.iIterCount
            * (data.iCols - 2) * (data.iRows - 2)
            / (data.fTimeStop - data.fTimeStart) << endl;
    }
    return;
}

// Initializes matrix
// Assumes exact solution is u(x,y) = (1-x^2)*(1-y^2)
void InitializeMatrix(JacobiData &data)
{
    /* Initilize initial condition and RHS */
    for (int j = data.iRowFirst; j <= data.iRowLast; j++)
    {
        for (int i = 0; i < data.iCols; i++)
        {
            // TODO: check if this values have to be ints or doubles
            int xx = (int) (-1.0 + data.fDx * i);
            int yy = (int) (-1.0 + data.fDy * j);

            int xx2 = xx * xx;
            int yy2 = yy * yy;

            U(j,i) = 0.0;
            F(j,i) = -data.fAlpha * (1.0 - xx2) * (1.0 - yy2)
                    + 2.0 * (-2.0 + xx2 + yy2);
        }
    }  
}

// Checks error between numerical and exact solution
void CheckError(JacobiData &data)
{
    double error = 0.0;

    for (int j = data.iRowFirst; j <= data.iRowLast; j++)
    {
        if ((data.iMyRank != 0 && j == data.iRowFirst) || 
            (data.iMyRank != data.iNumProcs - 1 && j == data.iRowLast)) continue;

        for (int i = 0; i < data.iCols; i++)
        {
            double xx   = -1.0 + data.fDx * i;
            double yy   = -1.0 + data.fDy * j;
            double temp = U(j,i) - (1.0 - xx * xx) * (1.0 - yy * yy);
            error += temp * temp;
        }
    }

    data.fError = sqrt(error) / (data.iCols * data.iRows);
}


int main (int argc, char** argv)
{
    int retVal = 0;    // return value

    JacobiData myData;


    // sets default values or reads from stdin
    // inits MPI and OpenMP if needed
    // distribute MPI data, calculate MPI bounds
    Init(myData, argc, argv);

    if (myData.afU && myData.afF)
    {
        // matrix init
        InitializeMatrix(myData);

        // starting timer
        myData.fTimeStart = GetRealTime();

        // running calculations
        Jacobi(myData);

        // stopping timer
        myData.fTimeStop = GetRealTime();

        // error checking
        CheckError(myData);

        // print result summary
        PrintResults(myData);
    }
    else
    {
        cout << " Memory allocation failed ...\n";
        retVal = -1;
    }

    // cleanup
    Finish(myData);
    return retVal;
}

