      subroutine jacobi (n,m,dx,dy,alpha,omega,u,f,tol,maxit)
******************************************************************
* Subroutine HelmholtzJ
* Solves poisson equation on rectangular grid assuming : 
* (1) Uniform discretization in each direction, and 
* (2) Dirichlect boundary conditions 
* 
* Jacobi method is used in this routine 
*
* Input : n,m   Number of grid points in the X/Y directions 
*         dx,dy Grid spacing in the X/Y directions 
*         alpha Helmholtz eqn. coefficient 
*         omega Relaxation factor 
*         f(n,m) Right hand side function 
*         u(n,m) Dependent variable/Solution
*         tol    Tolerance for iterative solver 
*         maxit  Maximum number of iterations 
*
* Output : u(n,m) - Solution 
*****************************************************************
      implicit none 
      integer n,m,maxit
      double precision dx,dy,f(n,m),u(n,m),alpha, tol,omega
*
* Local variables 
* 
      integer i,j,k,k_local 
      double precision resid,rsum,ax,ay,b
      double precision error_local, uold(n,m)
      double precision, ALLOCATABLE, DIMENSION(:) ::  error
      real ta,tb,tc,td,te,ta1,ta2,tb1,tb2,tc1,tc2,td1,td2
      real te1,te2
*
* Initialize coefficients 
      allocate (error(0:maxit)) !(0:maxit-1)
      ax = 1.0/(dx*dx) ! X-direction coef 
      ay = 1.0/(dy*dy) ! Y-direction coef
      b  = -2.0/(dx*dx)-2.0/(dy*dy) - alpha ! Central coeff
      k = 1
      error(k-1) = 10.0 * tol



      do while (k .le. maxit .and. error(k-1) .gt. tol)

         error(k) = 0.0

* Copy new solution into old
         do j=1,m
            do i=1,n
               uold(i,j) = u(i,j) 
            enddo
         enddo

* Compute stencil, residual, & update
         do j = 2,m-1
            do i = 2,n-1 

*     Evaluate residual 
               resid = (ax*(uold(i-1,j) + uold(i+1,j)) 
     &                + ay*(uold(i,j-1) + uold(i,j+1))
     &                + b * uold(i,j) - f(i,j))/b
*               resid = 2E+0*((ax*(uold(i-1,j) + uold(i+1,j)) 
*     &                + ay*(uold(i,j-1) + uold(i,j+1))
*     &                + b * uold(i,j) - f(i,j))/b)/2.0
* Update solution 
               u(i,j) = uold(i,j) - omega * resid
* Accumulate residual error
               error(k) = error(k) + resid*resid 
            end do
         enddo

* Error checK
         error(n*m) = dble(n*m)/sqrt(error(k))
         error(k) = 1 / error(k)
         WRITE (*,*) "Error: ", k, "--> ", error(k)
         k = k + 1

      enddo                     ! End iteration loop 
*
      print *, 'Total Number of Iterations ', k
      print *, 'Residual                   ', error(k-1) 
      deallocate(error) 
      return 
      end 
